/* OpenCA libpki package
* (c) 2000-2007 by Massimiliano Pala and OpenCA Group
* All Rights Reserved
*
* ===================================================================
* Released under OpenCA LICENSE
*/

#include <libpki/pki.h>

#define BUFF_MAX_SIZE	2048

/*! \brief Frees the memory assiociated with a PKI_HTTP data structure */

void PKI_HTTP_free ( PKI_HTTP *rv ) {
	if ( !rv ) return;

	if ( rv->location ) PKI_Free ( rv->location );
	if ( rv->type ) PKI_Free ( rv->type );

	if ( rv->body ) PKI_MEM_free ( rv->body );
	if ( rv->head ) PKI_MEM_free ( rv->head );

	PKI_Free ( rv );

	return;
}


/*! \brief Allocates the memory for a new PKI_HTTP data structure */

PKI_HTTP * PKI_HTTP_new ( void ) {

	PKI_HTTP *ret = NULL;

	if((ret = PKI_Malloc ( sizeof( PKI_HTTP ))) == NULL ) {
		return ( NULL );
	}

	/* Standard Header Codes */
	ret->code = 0;
	ret->type = NULL;
	ret->location = NULL;

	/* Data */
	ret->body = NULL;
	ret->head = NULL;

	return ( ret );
}

/*! \brief Returns a PKI_HTTP from the content of a char * */

char * PKI_HTTP_get_header_txt ( char *orig_data, char *header) {

	char *tk = NULL, *pnt = NULL;
	char *ret = NULL;

	char *data = NULL;
	int found = 0;

	if( !orig_data || !header || 
		(strlen(orig_data) == 0) || (strlen(header) == 0)) return NULL;

	if(( data = strdup( orig_data)) == NULL ) {
		PKI_log_debug("Memory Error");
		return NULL;
	}

	for ( tk = strtok_r ( data, "\r\n", &pnt ); tk;
			tk = strtok_r ( NULL, "\r\n", &pnt )) {

		if ( tk == NULL ) break;

		if ( strncmp_nocase( tk, header, strlen(header)) == 0 ) {
			found = 1;
			break;
		}
	}

	if ( found == 0 ) {
		PKI_Free ( data );
		return NULL;
	}

	if((pnt = strchr( tk, ':' )) == NULL ) {
		PKI_Free ( data );
		return NULL;
	}
	pnt++;

	while ( (pnt != NULL ) && (*pnt == ' ' ) ) {
			pnt++;
	};

	if( pnt ) {
		ret = strdup( pnt );
	}

	PKI_Free ( data );

	return ret;
}


/*! \brief Returns a PKI_HTTP from the content of a PKI_MEM */

char * PKI_HTTP_get_header ( PKI_HTTP *http, char * header ) {

	if( !http || !http->head || !header ) return NULL;

	return PKI_HTTP_get_header_txt ( (char *)http->head->data, header);
}

/* Internal version, can handle both HTTP and HTTPS */

PKI_HTTP *PKI_HTTP_get_message ( PKI_SOCKET *sock, int timeout, size_t max_size ) {

	PKI_HTTP * ret = NULL;

	PKI_MEM  * body = NULL;
	PKI_MEM  * header = NULL;

	char buf[1024];
	char * data = NULL;
	char * eoh = NULL;

	int read = 0;

	// size_t head_size = 0;
	// size_t data_size = 0;
	// size_t cont_size = 0;

	/*
	if ( max_size <= 2*sizeof(buf) ) {
		max = 64*1024 - sizeof(buf);
	} else {
		max = max_size - sizeof(buf);
	}

	if((data = PKI_Malloc ( max + sizeof(buf) )) == NULL ) {
		PKI_log_debug("Memory Error");
		return NULL;
	}
	*/

	if((ret = PKI_HTTP_new()) == NULL ) {
		PKI_log_debug("Memory Error");
		return NULL;
	}

	if(( header = PKI_MEM_new_null()) == NULL ) {
		PKI_ERROR(PKI_ERR_MEMORY_ALLOC, NULL );
		goto err;
	};

	if(( body = PKI_MEM_new_null()) == NULL ) {
		PKI_ERROR(PKI_ERR_MEMORY_ALLOC, NULL );
		goto err;
	};

	// data_size = 0;
	for ( ; ; ) {
		char *tmp_len = NULL;
		unsigned long msg_size = 0;

		if((read = PKI_SOCKET_read( sock, buf, sizeof(buf), timeout)) < 0) {
			PKI_ERROR(PKI_ERR_URI_READ, NULL );
			goto err;
		};

		// PKI_log_debug( "READ %d bytes", read );

		/* Fast And Ugly Hack to Speedup Data Read for HTTP */
		if ( read > 0 ) {
			// if ( data_size < 4 ) {
			// 	start = 0;
			// } else {
			// 	start = data_size - 4;
			// }

			// if( PKI_MEM_add ( data, buf, read ) == PKI_ERR ) {
			// 	PKI_ERROR(PKI_ERR_MEMORY_ALLOC, NULL );
			// 	goto err;
			// };

			/* Add the buff to the data */
			// memcpy(&data[data_size], buf, read );

			/* Update the Size of read data */
			// data_size += read;

			if( ret->head ) {
				if( PKI_MEM_add( body, buf, read ) == PKI_ERR ) {
					PKI_ERROR(PKI_ERR_MEMORY_ALLOC, NULL );
					goto err;
				};
			} else {
				// We have to figure out if we have the full head here,
				// if we do, we start the new body, otherwise we just add
				// the data to the header
				if ( ((eoh = strstr( buf, "\r\n\r\n")) != NULL) ||
						((eoh = strstr( buf, "\n\n")) != NULL ) ) {

					size_t grow_mem = 0;

					while ((*eoh != '\x0' ) && (*eoh == '\r' || *eoh == '\n')) {
						eoh++;
					};

					// Let's add the tail of the header to the header PKI_MEM
					if( eoh - buf > 0 ) {
						if( PKI_MEM_add ( header, buf, eoh - buf ) == PKI_ERR ) {
							PKI_ERROR(PKI_ERR_MEMORY_ALLOC, NULL );
							goto err;
						};
					};

					// Now let's add the header to the ret
					ret->head = header;

					// head_size = eoh - header->data;
					// ret->head = PKI_MEM_new_data( head_size + 1, 
					// 		(unsigned char *) header->data);
					// ret->head->data[head_size] = '\x0';

					// Now We create the new message with only the body of
					// the HTTP message
					grow_mem = &buf[read] - eoh;
					if(grow_mem > 0) {
						if(PKI_MEM_add(body, eoh, grow_mem) == PKI_ERR) {
							PKI_ERROR(PKI_ERR_MEMORY_ALLOC, NULL );
							goto err;
						};
					};

				} else {
					char *data = NULL;

					PKI_MEM_add(header, buf, read);

					data = (char *) header->data;

					if ( ((eoh = strstr( data, "\r\n\r\n")) != NULL) ||
							((eoh = strstr( data, "\n\n")) != NULL) ) {

						while ((*eoh != '\x0' ) && (*eoh == '\r' || *eoh == '\n')) {
							eoh++;
						};

						if ( eoh - data < header->size ) {
							// Part of the Body of the message has spilled in the
							// header, we have to remove it
							PKI_MEM_add(body, eoh, eoh - &data[header->size]);

							// Fix the header
							header->size = eoh - data;
							*eoh = '\x0';
						};

						// In any case, here we found the header
						ret->head = header;
					};
				};
			};
		}

		if(!tmp_len) {
			if((tmp_len = PKI_HTTP_get_header(ret, "Content-Length" )) != NULL ) {
				msg_size = atoll(tmp_len);
				PKI_Free(tmp_len);
			};
		};
	
		if(ret && body && (msg_size > 0) && (body->size >= msg_size))
			break;

		/*
		if( !ret->head ) {
			if ( (eoh = strstr( buf, "\r\n\r\n") ||
					eoh = strstr( buf, "\n\n") ) != NULL ) {

				char * pnt = NULL;
				PKI_MEM *tmpMessage = NULL;

				while ((*eoh != '\x0' ) && (*eoh == '\r' || *eoh == '\n')) {
					eoh++;
				};

				head_size = eoh - message->data;
				ret->head = PKI_MEM_new_data( head_size + 1, 
							(unsigned char *) message->data);
				ret->head->data[head_size] = '\x0';

				// Now We create the new message with only the body of
				// the HTTP message
				if(( tmpMessage = PKI_MEM_new_data ( eoh, 
						message->size - head_size )) == NULL ) {
					PKI_ERR(PKI_ERR_MEMORY_ALLOC, NULL );
					goto break;
				};

				// Free the old message
				PKI_Free ( message );

				// Switch the content with the new BODY-only message
				message = tmpMessage;

				if((pnt = PKI_HTTP_get_header(ret, "Content-Length")) 
						!= NULL) {
					cont_size = atoll(pnt);
					PKI_Free ( pnt );
				}
			}
		}
		*/
 
		/*
		// If we got headers and content-size
		if ( (ret->head) && (message->size > 0) ) {
			if ( cont_size <= data_size - head_size ) break;
		}

		// Error, or end of data
		if ( read <= 0 ) {
			// PKI_log_debug("FINISHED READING => %d + %d (%d)",
			// 	head_size, cont_size, data_size );

			cont_size = data_size - head_size;
			break;
		}
		*/

		if ( read <= 0 ) break;
	}

	if( ret->head ) {
		if(strncmp_nocase((char *)ret->head->data, PKI_HTTP_METHOD_GET_TXT, 
					strlen(PKI_HTTP_METHOD_GET_TXT)) == 0) {
			ret->method = PKI_HTTP_METHOD_GET;
			ret->url = URL_new ( &data[4] );
		} else if (strncmp_nocase((char *)ret->head->data, PKI_HTTP_METHOD_POST_TXT,
					strlen(PKI_HTTP_METHOD_POST_TXT)) == 0) {
			ret->method = PKI_HTTP_METHOD_POST;
		} else if (strncmp_nocase((char *)ret->head->data, "HTTP/", 5) == 0 ) {
			if(sscanf((char *) ret->head->data,"HTTP/%f %d", 
						&ret->version, &ret->code) < 1) {
				PKI_log_debug("ERROR Parsing HTTP MSG");
				PKI_HTTP_free(ret);
				return NULL;
			}
		} else {
			PKI_log_debug("ERROR::Malformed HTTP message => %s", ret->head->data);
			goto err;
		}
	} else {
		PKI_log_debug("ERROR::Missing header in HTTP message");
		goto err;
	};

	if ( body->size == 0 ) {
		PKI_log_debug( "WARNING: No content!");
	} else {
		ret->body = body;
		/*
		ret->body = PKI_MEM_new_data ( cont_size, 
				(unsigned char *) &data[data_size-cont_size] );
		PKI_log_debug("HTTP DATA => size (%d->%d)", 
			data_size-cont_size, cont_size );
		*/
	}
 
	ret->location = PKI_HTTP_get_header ( ret, "Location" );
	ret->type = PKI_HTTP_get_header ( ret, "Content-Type" );

	/* Free the buffer memory */
	// PKI_Free ( data );

	return ( ret );

err:

	if ( body && ( !ret || !ret->body ) ) {
		PKI_MEM_free ( body );
		body = NULL; // Avoid double free when freeing ret
	};

	if ( header && (!ret || !ret->head) ) {
		PKI_MEM_free ( header );
		header = NULL; // Avoid double free when freeing ret
	};

	if ( ret ) PKI_HTTP_free(ret);

	return NULL;
}

/*! \brief Sends a HTTP message to a URL and retrieve the response
 *
 * Sends (POST/GET) data to a url and (if a pointer to a mem stack
 * is provided) returns the received response. PKI_ERR is returned
 * in case of error, otherwise PKI_OK is returned.
 */

int PKI_HTTP_get_url (URL *url, char *data, size_t data_size,
			char *content_type, int method, int timeout, 
			size_t max_size, PKI_MEM_STACK **sk, PKI_SSL *ssl ) {

	PKI_SOCKET *sock = NULL;
	int ret = 0;

	if (!url) return PKI_ERR;

	sock = PKI_SOCKET_new();
	if ( ssl ) {
		PKI_SOCKET_set_ssl ( sock, ssl );
	}

	if ( PKI_SOCKET_open_url ( sock, url, timeout ) == PKI_ERR ) {
		PKI_SOCKET_free ( sock );
		return PKI_ERR;
	}

	ret = PKI_HTTP_get_socket ( sock, data, data_size, content_type,
				method, timeout, max_size, sk );

	PKI_SOCKET_close ( sock );
	PKI_SOCKET_free ( sock );

	return ret;
}

/*! \brief Reads a data from an HTTP server */

int PKI_HTTP_get_socket (PKI_SOCKET *sock, char *data, size_t data_size,
			char *content_type, int method, int timeout, 
			size_t max_size, PKI_MEM_STACK **sk ) {

	size_t len          = 0;

	char *my_cont_type	= "application/unknown";

	PKI_HTTP *http_rv	= NULL;

	int	      rv		= -1;
	int 	  ret		= PKI_OK;

	size_t max_len = 0;

	char *tmp  = NULL;
    
	char *head_get =
			"GET %s HTTP/1.1\r\n"
			"Host: %s\r\n"
			"User-Agent: LibPKI\r\n"
	        "Connection: close\r\n";

	char *head_post = 
			"POST %s HTTP/1.1\r\n"
			"Host: %s\r\n"
			"User-Agent: LibPKI\r\n"
			"Connection: close\r\n"
			"Content-type: %s\r\n"
			"Content-Length: %d\r\n";

	char *head = NULL;

	if ( timeout < 0 ) timeout = 0;

	if ( !sock || !sock->url ) return PKI_ERR;

	if ( method == PKI_HTTP_METHOD_GET ) {
		head = head_get;
		max_len = strlen(head) + strlen( sock->url->path ) + 
						strlen( sock->url->addr ) + 1;
		tmp = PKI_Malloc ( max_len );
		len = (size_t) snprintf(tmp, max_len, head, sock->url->path, 
						sock->url->addr );
	} else if ( method == PKI_HTTP_METHOD_POST ) {
		head = head_post;
		if ( content_type ) my_cont_type = content_type;
		max_len = strlen(head) + strlen( sock->url->path ) + 
				strlen( sock->url->addr ) + strlen(my_cont_type) + 101;
		tmp = PKI_Malloc ( max_len );
		len = (size_t) snprintf(tmp, max_len, head, sock->url->path, sock->url->addr, 
					my_cont_type, data_size );
	} else {
		PKI_log_err ( "Method (%d) not supported!", method );
		return PKI_ERR;
	}

	if((rv = PKI_SOCKET_write ( sock, tmp, len )) < 0 ) {
			PKI_log_err("Can not write HTTP header to socket");
			goto err;
	}
	PKI_Free ( tmp );

	if ( sock->url->usr && sock->url->pwd ) {

		tmp = PKI_Malloc ( strlen(sock->url->usr) + strlen(sock->url->pwd) + 100);
		sprintf( tmp, "Authentication: user %s:%s\r\n\r\n", 
							sock->url->usr, sock->url->pwd);
	} else {
		tmp = strdup ("\r\n");
	}
	len = strlen( tmp );

	if((rv=PKI_SOCKET_write ( sock, tmp, len)) < 0) {
		/* Error in write */
		PKI_log_err("Can not write HTTP header to socket");
		goto err;
	}

	if( (method == PKI_HTTP_METHOD_POST) && ( data != NULL ) ) {
		if((rv = PKI_SOCKET_write( sock, data, data_size)) < 0) {
			PKI_log_err ("Can not write POST to socket.");
			goto err;
		}
	}
	
	if((http_rv = PKI_HTTP_get_message ( sock, timeout, max_size))== NULL){
		PKI_log_err ("HTTP retrieval error");
		goto err;
	}

	if ( http_rv->code >= 400 ) {
		goto err;
	} else if ( http_rv->code >= 300 ) {
		/* Redirection - let's try that */

		if( !http_rv->location ) {
			PKI_log_debug ( "HTTP Redirection but no location "
						"provided!");
			goto err;
		}

		PKI_log_debug("HTTP Redirection Location ==> %s", 
							http_rv->location );

		if( strstr(http_rv->location, "://") != NULL ) {
			URL *url_tmp = NULL;

			if( strncmp_nocase( http_rv->location, sock->url->url_s, 
					(int) strlen(http_rv->location)) == 0) {
				PKI_log_debug( "HTTP cyclic redirection!");
				goto err;
			}

			if(( url_tmp = URL_new ( http_rv->location )) == NULL) {
				PKI_log_debug("HTTP location is not a valid "
					"URI (%s)", http_rv->location );
				goto err;
			}

			if ( sock->url->ssl == 0 ) {
				ret = PKI_HTTP_get_url ( url_tmp, data, 
					data_size, content_type, method, timeout, 
							max_size, sk, NULL );
			} else {
				PKI_SSL *ssl2 = PKI_SSL_dup ( sock->ssl );

				ret = PKI_HTTP_get_url ( url_tmp, data, 
					data_size, content_type, method, timeout, 
							max_size, sk, ssl2 );
			}

			if ( url_tmp ) URL_free ( url_tmp );
	
			goto end;

		} else {
			const char *prot_s = NULL;
			char new_url[2048];
			URL *my_new_url = NULL;
			PKI_SSL *ssl2 = PKI_SSL_dup ( sock->ssl );

			prot_s = URL_proto_to_string ( sock->url->proto );
			if( !prot_s ) goto err;

			snprintf(new_url, sizeof(new_url),"%s://%s%s",
				prot_s, sock->url->addr, http_rv->location );

			if( strncmp_nocase( new_url, sock->url->url_s, 
					(int) strlen ( new_url )) == 0 ) {
				PKI_log_debug( "HTTP cyclic redirection!");
				goto err;
			}

			my_new_url = URL_new ( new_url );

			ret = PKI_HTTP_get_url ( my_new_url, data, 
					data_size, content_type, method, 
						timeout, max_size, sk, ssl2 );

			if (ssl2) PKI_SSL_free ( ssl2 );
		}
	} else if ( http_rv->code != 200 ) {
		PKI_log_debug( "HTTP Return code not manageable (%d)",
							http_rv->code );
		goto err;
	}

	/* If a Pointer was provided, we want the data back :D */
	if ( sk ) {
		if ( ! *sk ) {
			if((*sk = PKI_STACK_MEM_new()) == NULL ) {
				PKI_log_debug ("Memory Failure");
				goto err;
			}
		}

		/* Add the returned value to the stack */
		PKI_STACK_MEM_push ( *sk, PKI_MEM_dup(http_rv->body));
	}

end:
	if ( http_rv ) PKI_HTTP_free ( http_rv );

	return ret;

err:
	if ( http_rv ) PKI_HTTP_free ( http_rv );

	return PKI_ERR;
}

/* ------------------------------- HTTP GET --------------------------- */

/*! \brief Returns the data from an HTTP source by using the GET command */

int PKI_HTTP_GET_data ( char *url_s, int timeout, size_t max_size,
					PKI_MEM_STACK **ret, PKI_SSL *ssl ) {

	URL *url = NULL;
	int rv = PKI_OK;

	if( !url_s ) return PKI_ERR;

	if((url = URL_new( url_s )) == NULL ) {
		return PKI_ERR;
	}

	rv = PKI_HTTP_get_url ( url, NULL, 0, NULL,
			PKI_HTTP_METHOD_GET, timeout, max_size, ret, ssl );

	if ( url ) URL_free ( url );
	return rv;
}

/*! \brief Returns the data from an HTTP URL by using the GET command */

int PKI_HTTP_GET_data_url ( URL *url, int timeout, size_t max_size,
					PKI_MEM_STACK **ret, PKI_SSL *ssl ) {

	if ( !url ) return PKI_ERR;

	return PKI_HTTP_get_url ( url, NULL, 0, NULL,
			PKI_HTTP_METHOD_GET, timeout, max_size, ret, ssl );
}

/*! \brief Returns HTTP data from a PKI_SOCKET by using the GET command */

int PKI_HTTP_GET_data_socket ( PKI_SOCKET *sock, int timeout, size_t max_size,
					PKI_MEM_STACK **ret ) {

	if ( !sock ) return PKI_ERR;

	return PKI_HTTP_get_socket ( sock, NULL, 0, NULL,
			PKI_HTTP_METHOD_GET, timeout, max_size, ret );
}

/* ------------------------------- HTTP POST --------------------------- */

/*! \brief Use POST method to transfer data via HTTP. If a pointer to a
 *         PKI_MEM_STACK (eg., &stack) is provided, the returned data isi
 *         added to it
 */

int PKI_HTTP_POST_data ( char *url_s, char *data, size_t size, 
			char *content_type, int timeout, size_t max_size,
				PKI_MEM_STACK **ret_sk, PKI_SSL *ssl ) {

	URL *url = NULL;
	int ret = PKI_OK;

	if( !url_s || !data || !content_type ) {
		/* ERROR: All data are strictly required! */
		return PKI_ERR;
	}

	if((url = URL_new(url_s)) == NULL ) {
		/* Error in creating the URL structure */
		return PKI_ERR;
	}

	ret = PKI_HTTP_get_url ( url, data, size, content_type,
			PKI_HTTP_METHOD_POST, timeout, max_size, ret_sk, ssl);

	if ( url ) URL_free (url);

	return ret;
}

/*! \brief Use POST method to transfer data via HTTP. If a pointer to a
 *         PKI_MEM_STACK (eg., &stack) is provided, the returned data is
 *         added to it
 */

int PKI_HTTP_POST_data_url ( URL *url, char *data, size_t size, 
			char *content_type, int timeout, size_t max_size,
				PKI_MEM_STACK **ret_sk, PKI_SSL *ssl ) {


	if ( !url ) return PKI_ERR;

	return PKI_HTTP_get_url ( url, data, size, content_type,
			PKI_HTTP_METHOD_POST, timeout, max_size, ret_sk, ssl);
}

/*! \brief Use POST method to transfer data via HTTP. If a pointer to a
 *         PKI_MEM_STACK (eg., &stack) is provided, the returned data is
 *         added to it
 */

int PKI_HTTP_POST_data_socket ( PKI_SOCKET *sock, char *data, size_t size, 
			char *content_type, int timeout, size_t max_size,
				PKI_MEM_STACK **ret_sk ) {


	if ( !sock ) return PKI_ERR;

	return PKI_HTTP_get_socket ( sock, data, size, content_type,
			PKI_HTTP_METHOD_POST, timeout, max_size, ret_sk );
}

