/* PKI_X509 object management */

#include <libpki/pki.h>

/* Static function for managing read callbacks for auto format downloading */

static void * __get_data_callback ( PKI_IO *io, const PKI_X509_CALLBACKS *cb,
				PKI_DATA_FORMAT format, PKI_CRED *cred );
/*
static PKI_MEM *__keypair_put_mem_value (EVP_PKEY *x, PKI_DATATYPE type,
                        PKI_MEM **pki_mem, PKI_DATA_FORMAT format,
                                PKI_CRED *cred, HSM *hsm);
*/

/* -------------------------- X509 mem Operations -------------------- */

/*! \brief Reads a PKI_X509 object from a PKI_MEM */

PKI_X509 *PKI_X509_get_mem ( PKI_MEM *mem, PKI_DATATYPE type, 
						PKI_CRED *cred, HSM *hsm ) {

	PKI_X509_STACK *sk = NULL;
	PKI_X509 *ret = NULL;

	if((sk = PKI_X509_STACK_get_mem ( mem, type, cred, hsm )) == NULL ) {
		return NULL;
	}

	ret = PKI_STACK_X509_pop ( sk );
	PKI_STACK_X509_free ( sk );

	return ret;
}

void * PKI_X509_get_mem_value ( PKI_MEM *mem, PKI_DATATYPE type,
						PKI_CRED *cred, HSM *hsm ) {
	PKI_X509 *tmp = NULL;
	void *ret = NULL;

	if(( tmp = PKI_X509_get_mem ( mem, type, cred, hsm )) == NULL ) {
		return NULL;
	}

	if ( tmp->value ) {
		ret = PKI_X509_dup_value ( tmp );
	}

	PKI_X509_free ( tmp );

	return ret;
}


static void * __get_data_callback ( PKI_IO *io, const PKI_X509_CALLBACKS *cb,
				PKI_DATA_FORMAT format, PKI_CRED *cred ) {
	void *ret = NULL;
	char *pwd = NULL;

	if (!io || !cb ) return NULL;

	if ( cred && cred->password ) pwd = (char *) cred->password;

	switch ( format ) {
		case PKI_DATA_FORMAT_PEM :
			if( cb->read_pem ) {
				ret = cb->read_pem (io, NULL, NULL, pwd );
			}
			break;
		case PKI_DATA_FORMAT_ASN1 :
			if( cb->read_der ) {
				ret = cb->read_der ( io, NULL );
			}
			break;
		case PKI_DATA_FORMAT_TXT :
			if ( cb->read_txt ) {
				ret = cb->read_txt ( io, NULL );
			}
			break;
		case PKI_DATA_FORMAT_B64 :
			if ( cb->read_b64 ) {
				ret = cb->read_b64 ( io, NULL );
			} else if ( cb->read_der ) {
				PKI_MEM *rr = NULL;

				rr = PKI_MEM_new_bio ( io, NULL );
				PKI_MEM_B64_decode ( rr );
				(void) BIO_flush(io);
				BIO_write ( io, rr->data, rr->size );
				ret = cb->read_der ( io, NULL );
				PKI_MEM_free ( rr );
			};
			break;
		case PKI_DATA_FORMAT_XML :
			if ( cb->read_xml ) {
				ret = cb->read_xml ( io, NULL );
			}
			break;
		default:
			return NULL;
	}

	return ret;
}

/*! \brief Returns a stack of objects read from the passed PKI_MEM */

PKI_X509_STACK *PKI_X509_STACK_get_mem ( PKI_MEM *mem, 
			PKI_DATATYPE type, PKI_CRED *cred, HSM *hsm ) {

	PKI_X509_STACK *sk = NULL;
	PKI_X509 * x_obj = NULL;
	const PKI_X509_CALLBACKS *cb = NULL;

	BUF_MEM *p = NULL;
	BIO *membio = NULL;

	char *temp = NULL;
	int cont = 1;
	int i = 0;
	long curr = 0;

	size_t mem_size = 0;
	char * mem_buf = NULL;

	if( !mem || mem->size <= 0 ) return (NULL);

	if((cb = PKI_X509_CALLBACKS_get ( type, hsm )) == NULL ) {
		return NULL;
	}

	if((temp = strstr((char *) mem->data, "-----BEGIN" )) == NULL ) {
		mem_buf = (char *) mem->data;
		mem_size = mem->size;
	} else {
		mem_buf = temp;
		mem_size = (size_t) ((char *)mem->data - temp) + mem->size;
	}

	if((membio = BIO_new_mem_buf( mem_buf, (int) mem_size )) == NULL ) {
		return( NULL );
	}

	if(BIO_set_close(membio, BIO_NOCLOSE) != 1 ) {
		BIO_free (membio);
		return(NULL);
	}

	if((sk = PKI_STACK_X509_new()) == NULL ) {
		BIO_free( membio );
		return(NULL);
	}

	cont = 1;
	p = (BUF_MEM *) membio->ptr;
	curr = p->length;
	while ((cont == 1) && (p->length > 0)) {

		curr = p->length;

		/* Check we have a good type */
		if((x_obj = PKI_X509_new ( type, hsm )) == NULL )
			break;

		/* We need the callbacks -- make sure they are there! */
		if ( x_obj->cb == NULL ) break;

		for ( i = 1; i < PKI_DATA_FORMAT_SIZE; i++ ) {
			curr = p->length;

			x_obj->value = __get_data_callback ( membio, cb, 
						i, cred );
			if( x_obj->value == NULL ) {
				p->length = curr;
				p->data -= curr;
			} else {
				break;
			}
		}

		if ( x_obj->value == NULL ) {
			PKI_X509_free ( x_obj );
			cont = 0;
			break;
		} else {
			if ( cred ) x_obj->cred = PKI_CRED_dup ( cred );

			PKI_STACK_X509_push( sk, x_obj );
		}
	}

	if(membio) BIO_free( membio );
	return( sk );
}

/*! \brief Writes a PKI_X509 object to a PKI_MEM structure */

PKI_MEM *PKI_X509_put_mem (PKI_X509 *x, PKI_DATA_FORMAT format, 
				PKI_MEM **mem, PKI_CRED *cred ) {

	PKI_DATATYPE type;

	if( !x || !x->value ) {
		PKI_ERROR(PKI_ERR_PARAM_NULL, NULL);
		return ( NULL );
	};

	if( !x->cb ) {
		PKI_ERROR(PKI_ERR_CALLBACK_NULL, NULL);
		return NULL;
	};

	type = PKI_X509_get_type ( x );

	PKI_X509_set_modified ( x );

	return PKI_X509_put_mem_value ( x->value, type, mem, 
					format, cred, x->hsm );
}

/*! \brief Writes a PKI_X509_XXX_VALUE to a PKI_MEM structure */

PKI_MEM *PKI_X509_put_mem_value (void *x, PKI_DATATYPE type, 
			PKI_MEM **pki_mem, PKI_DATA_FORMAT format, 
				PKI_CRED *cred, HSM *hsm) {

	PKI_IO *membio = NULL;
	const PKI_X509_CALLBACKS *cb = NULL;
	PKI_MEM *ret = NULL;
	int rv = 0;

	char *pwd = NULL;
	const EVP_CIPHER *enc = NULL;

	if(( membio = BIO_new(BIO_s_mem())) == NULL ) {
		PKI_ERROR(PKI_ERR_OBJECT_CREATE, NULL);
		return NULL;
	}

	if(( cb = PKI_X509_CALLBACKS_get ( type, hsm )) == NULL ) {
		PKI_ERROR(PKI_ERR_CALLBACK_NULL, NULL);
		return NULL;
	}

	/* Check if we have to encrypt the key */
	if ( cred && cred->password && strlen(cred->password) > 0 ) {
		pwd = (char *) cred->password;
		enc=EVP_aes_256_cbc();
	} else {
		enc=NULL;
	};

	switch( format ) {
		case PKI_DATA_FORMAT_PEM:
			if( cb->to_pem_ex ) {
			  // if( type == PKI_DATATYPE_X509_KEYPAIR ) {
				// rv = PEM_write_bio_PrivateKey( membio,
				//	x, enc, NULL, 0, NULL, pwd);
				rv = cb->to_pem_ex(membio, x, (void *) enc, NULL, 0, NULL, pwd );
			} else if( cb->to_pem ) {
				rv = cb->to_pem ( membio, x );
			}
			break;

		case PKI_DATA_FORMAT_URL:
		case PKI_DATA_FORMAT_ASN1:
			if( cb->to_der ) {
				rv = cb->to_der ( membio, x );
			} else {
				PKI_log_debug ( "NO ASN1 (type %d) callback ? %p",
					type, cb->to_der );
			}
			break;

		case PKI_DATA_FORMAT_TXT:
			if( cb->to_txt ) {
				rv = cb->to_txt ( membio, x );
			}
			break;

		case PKI_DATA_FORMAT_B64:
			if( cb->to_b64 ) {
				rv = cb->to_b64 ( membio, x );
			} else if ( cb->to_der ) {
				rv = cb->to_der ( membio, x );
				if((ret = PKI_MEM_new_bio(membio, pki_mem)) 
								!= NULL ) {
					rv = PKI_MEM_B64_encode ( ret );
				} else {
					rv = 0;
				}
			}
			break;

		case PKI_DATA_FORMAT_XML:
			if( cb->to_xml ) {
				rv = cb->to_xml ( membio, x );
			}
			break;

		case PKI_DATA_FORMAT_UNKNOWN:
		default:
			PKI_log_debug("PKI_X509_put_mem_value()::Unsupported "
				"coding Format %d", format );
			rv = 0;
	}

	/* We already covered the case of B64 without a specific encoding callback */
	if ( (format != PKI_DATA_FORMAT_B64) || ( cb->to_b64 != NULL ) ) {
		if ( rv ) {
			if((ret = PKI_MEM_new_bio(membio, pki_mem)) == NULL ) {
				PKI_log_err ("Can not convert BIO");
				if ( membio ) BIO_free_all ( membio );
				return NULL;
			}
		} else {
			/* OpenSSL ERROR! */
			if( membio ) BIO_free_all ( membio );
			return ( NULL );
		}
	}

	BIO_free_all ( membio );

	if( ret && (format == PKI_DATA_FORMAT_URL)) {
		PKI_MEM *url_encoded = NULL;
	
		fprintf(stderr, "DEBUG::: FORMAT IS URL!!!!\n");

		if((url_encoded = PKI_MEM_url_encode(ret, 1 )) == NULL ) {
			PKI_MEM_free(ret);
			return NULL;
		};
		fprintf(stderr, "DEBUG::: BIN ENCODED, SIZE => %d !!!!\n", ret->size );
		fprintf(stderr, "DEBUG::: URL ENCODED, SIZE => %d !!!!\n", url_encoded->size );
		fprintf( stderr, "DEBUG::ENCODED => %s\n\n", url_encoded->data);

		// PKI_MEM_free(ret);
		ret = url_encoded;
	};

	return ret;
}

/*! \brief Writes a stack of PKI_X509 to a PKI_MEM */

PKI_MEM * PKI_X509_STACK_put_mem ( PKI_X509_STACK *sk, PKI_DATA_FORMAT format, 
				PKI_MEM **pki_mem, PKI_CRED *cred, HSM *hsm ) {

	PKI_MEM *ret = NULL;
	int i = 0;

	if( !sk ) return (PKI_ERR);

	if ( pki_mem != NULL ) {
		if( *pki_mem == NULL ) {
			if((*pki_mem = PKI_MEM_new_null()) == NULL ) {
				PKI_ERROR(PKI_ERR_MEMORY_ALLOC, NULL);
				return NULL;
			}
			ret = *pki_mem;
		} else {
			ret = *pki_mem;
		}
	} else {
		ret = PKI_MEM_new_null();
	}

	if ( !ret ) return NULL;

	for( i = 0; i < PKI_STACK_X509_elements ( sk ); i++ ) {
		PKI_X509 *x_obj = NULL;
		PKI_MEM *tmp_mem = NULL;

		if((x_obj = PKI_STACK_X509_get_num( sk, i )) != NULL) {

			if((tmp_mem = PKI_X509_put_mem ( x_obj, 
					format, pki_mem, cred )) == NULL ) {
				PKI_log_debug("ERROR adding item %d to PKI_MEM", i);
				continue;
			}
			if ( x_obj->cred ) PKI_CRED_free ( x_obj->cred );
			if ( cred ) x_obj->cred = PKI_CRED_dup ( cred );
		}
	}

	return ret;
}

